"""Functions for reducing arctans."""

from functools import cache
import math
from arctans.primes import is_gaussian_prime, complex_factorise
from arctans.arctans import arccotan, arctan, Zero, AbstractTerm, Arctan
from arctans.numbers import j


@cache
def _convert_rational_single_arctan(a: Arctan) -> AbstractTerm:
    """Convert a rational arccotangent into a sum of integral arccotangents.

    Args:
        a: An arctan

    Returns:
        A sum of integral arccotangents
    """
    if a.terms[0][1].numerator == 1:
        return a
    beta = int(a.terms[0][1].numerator)
    alpha = int(a.terms[0][1].denominator)

    out = Zero()
    sign = 1
    while beta > 0:
        n = alpha // beta
        alpha, beta = alpha * n + beta, alpha % beta
        assert isinstance(a.terms[0][0] * arccotan(n), AbstractTerm)
        out += sign * a.terms[0][0] * arccotan(n)
        sign *= -1
    return out


def convert_rational(a: AbstractTerm) -> AbstractTerm:
    """Convert a rational arccotangent into a sum of integral arccotangents.

    Args:
        a: An arctan or sum of arctans

    Returns:
        A sum of integral arccotangents
    """
    if isinstance(a, Arctan):
        return _convert_rational_single_arctan(a)
    out = Zero()
    for c, arg in a.terms:
        at = arctan(arg)
        assert isinstance(at, Arctan)
        out += c * _convert_rational_single_arctan(at)
    return out


@cache
def _reduce_single_arctan(a: Arctan) -> AbstractTerm:
    """Express an arctan as a sum of irreducible integral arccotangents.

    Args:
        a: An arctan

    Returns:
        A sum of irreducible integral arccotangents
    """
    n = a.terms[0][1].denominator + j * a.terms[0][1].numerator
    if is_gaussian_prime(n):
        return a
    out = Zero()
    for f in complex_factorise(n):
        out += a.terms[0][0] * convert_rational(arctan(f.imag / f.real))

    return out


def reduce(a: AbstractTerm) -> AbstractTerm:
    """Express an arctan as a sum of irreducible integral arccotangents.

    Args:
        a: An arctan or sum of arctans

    Returns:
        A sum of irreducible integral arccotangents
    """
    out = Zero()
    for c, arg in a.terms:
        at = arctan(arg)
        assert isinstance(at, Arctan)
        out += c * _reduce_single_arctan(at)

    k = int(math.floor((float(a) - float(out)) * 4 / math.pi + 0.1))
    out += k * arctan(1)
    return out



"""Numbers."""

from __future__ import annotations
import math
from abc import abstractmethod
from typing import Any
from arctans.core import Representable

try:
    from typing import Self  # type: ignore
except ImportError:
    from typing_extensions import Self


class AbstractNumber(Representable):
    """Base class for number."""

    @property
    @abstractmethod
    def real(self) -> RealNumber:
        """Real part."""

    @property
    @abstractmethod
    def imag(self) -> RealNumber:
        """Imaginary part."""

    @property
    @abstractmethod
    def numerator(self) -> AbstractNumber:
        """Numerator."""

    @property
    @abstractmethod
    def denominator(self) -> Integer:
        """Denominator."""

    @abstractmethod
    def conjugate(self) -> AbstractNumber:
        """Compute the complex conjugate."""

    @abstractmethod
    def __int__(self) -> int:
        pass

    @abstractmethod
    def __float__(self) -> float:
        pass

    @abstractmethod
    def __complex__(self) -> complex:
        pass

    @abstractmethod
    def _to_same_type(self, other: Any) -> Self:
        """Convert other to the same type as self."""

    @abstractmethod
    def _add(self, other: Self) -> AbstractNumber:
        """Add something of the same type to this."""

    @abstractmethod
    def _sub(self, other: Self) -> AbstractNumber:
        """Subtract something of the same type from this."""

    @abstractmethod
    def _mul(self, other: Self) -> AbstractNumber:
        """Multiply something of the same type by this."""

    @abstractmethod
    def _truediv(self, other: Self) -> AbstractNumber:
        """Divide this by something of the same type."""

    def _mod(self, other: Self) -> AbstractNumber:
        """Find the remainder when dividing this by something of the same type."""
        return NotImplemented

    def _floordiv(self, other: Self) -> AbstractNumber:
        """Find the remainder when dividing this by something of the same type."""
        return NotImplemented

    @abstractmethod
    def _eq(self, other: Self) -> bool:
        """Check if something of the same type is equal to this."""

    def _pow(self, other: int) -> AbstractNumber:
        """Raise to an integer power."""
        if other == 0:
            return Integer(1)
        elif other > 0:
            return _simplify_type(self._pow(other - 1) * self)
        else:  # other < 0
            return _simplify_type(self._pow(other + 1) / self)

    def __add__(self, other: Any):
        try:
            s, o = _as_common_type(self, other)
            return _simplify_type(s._add(o))
        except ValueError:
            return NotImplemented

    def __radd__(self, other: Any):
        try:
            s, o = _as_common_type(self, other)
            return _simplify_type(o._add(s))
        except ValueError:
            return NotImplemented

    def __sub__(self, other: Any):
        try:
            s, o = _as_common_type(self, other)
            return _simplify_type(s._sub(o))
        except ValueError:
            return NotImplemented

    def __rsub__(self, other: Any):
        try:
            s, o = _as_common_type(self, other)
            return _simplify_type(o._sub(s))
        except ValueError:
            return NotImplemented

    def __mul__(self, other: Any):
        try:
            s, o = _as_common_type(self, other)
            return _simplify_type(s._mul(o))
        except ValueError:
            return NotImplemented

    def __rmul__(self, other: Any):
        try:
            s, o = _as_common_type(self, other)
            return _simplify_type(o._mul(s))
        except ValueError:
            return NotImplemented

    def __neg__(self):
        return _simplify_type(-1 * self)

    def __truediv__(self, other: Any):
        try:
            s, o = _as_common_type(self, other)
            return _simplify_type(s._truediv(o))
        except ValueError:
            return NotImplemented

    def __rtruediv__(self, other: Any):
        try:
            s, o = _as_common_type(self, other)
            return _simplify_type(o._truediv(s))
        except ValueError:
            return NotImplemented

    def __mod__(self, other: Any):
        try:
            s, o = _as_common_type(self, other)
            return _simplify_type(s._mod(o))
        except ValueError:
            return NotImplemented

    def __rmod__(self, other: Any):
        try:
            s, o = _as_common_type(self, other)
            return _simplify_type(o._mod(s))
        except ValueError:
            return NotImplemented

    def __floordiv__(self, other: Any):
        try:
            s, o = _as_common_type(self, other)
            return _simplify_type(s._floordiv(o))
        except ValueError:
            return NotImplemented

    def __rfloordiv__(self, other: Any):
        try:
            s, o = _as_common_type(self, other)
            return _simplify_type(o._floordiv(s))
        except ValueError:
            return NotImplemented

    def __eq__(self, other) -> bool:
        try:
            s, o = _as_common_type(self, other)
            return s._eq(o)
        except ValueError:
            return NotImplemented

    def __pow__(self, other):
        try:
            return _simplify_type(self._pow(int(other)))
        except ValueError:
            return NotImplemented

    def __rpow__(self, other):
        try:
            return _simplify_type(other ** int(self))
        except ValueError:
            return NotImplemented

    def __iadd__(self, other):
        return self + other

    def __isub__(self, other):
        return self - other

    def __imul__(self, other):
        return self * other

    def __abs__(self):
        return math.sqrt(self.real**2 + self.imag**2)

    def __hash__(self):
        return hash(self.__repr__())


class RealNumber(AbstractNumber):
    """A real number."""

    @property
    def real(self) -> RealNumber:
        return self

    @property
    def imag(self) -> RealNumber:
        return Integer(0)

    def conjugate(self) -> AbstractNumber:
        return self

    def __lt__(self, other) -> bool:
        return int((self - other).numerator) < 0

    def __le__(self, other) -> bool:
        return int((self - other).numerator) <= 0

    def __gt__(self, other) -> bool:
        return int((self - other).numerator) > 0

    def __ge__(self, other) -> bool:
        return int((self - other).numerator) >= 0


class Integer(RealNumber):
    """An integer."""

    def __init__(self, i: int):
        """Initialise.

        Args:
            i: The integer
        """
        self._i = i

    def __str__(self):
        return f"{self._i}"

    def __repr__(self):
        return f"Integer({self._i})"

    def as_latex(self) -> str:
        return f"{self._i}"

    @property
    def numerator(self) -> AbstractNumber:
        return self

    @property
    def denominator(self) -> Integer:
        return Integer(1)

    def __int__(self) -> int:
        return self._i

    def __float__(self) -> float:
        return float(self._i)

    def __complex__(self) -> complex:
        return self._i + 0j

    def _to_same_type(self, other: Any) -> Integer:
        if isinstance(other, int):
            return Integer(other)
        if isinstance(other, Integer):
            return other
        raise ValueError(f"Could not convert {other} to integer")

    def _add(self, other: Self) -> AbstractNumber:
        return Integer(self._i + other._i)

    def _sub(self, other: Self) -> AbstractNumber:
        return Integer(self._i - other._i)

    def _mul(self, other: Self) -> AbstractNumber:
        return Integer(self._i * other._i)

    def _truediv(self, other: Self) -> AbstractNumber:
        return Rational(self._i, other._i)

    def _pow(self, other: int) -> AbstractNumber:
        return Integer(self._i**other)

    def _mod(self, other: Self) -> AbstractNumber:
        return Integer(self._i % other._i)

    def _floordiv(self, other: Self) -> AbstractNumber:
        return Integer(self._i // other._i)

    def _eq(self, other: Self) -> bool:
        return self._i == other._i

    def __abs__(self):
        return abs(self._i)

    def __hash__(self):
        return hash(self.__repr__())


class Rational(RealNumber):
    """A rational number."""

    def __init__(self, numerator: int, denominator: int):
        """Initialise.

        Args:
            numerator: The numerator
            denominator: The denominator
        """
        if denominator < 0:
            numerator *= -1
            denominator *= -1
        hcf = math.gcd(abs(numerator), abs(denominator))
        self._num = numerator // hcf
        self._den = denominator // hcf

    def as_latex(self) -> str:
        return f"\\frac{{{self._num}}}{{{self._den}}}"

    def __str__(self):
        return f"{self._num}/{self._den}"

    def __repr__(self):
        return f"Rational({self.__str__()})"

    @property
    def numerator(self) -> AbstractNumber:
        return Integer(self._num)

    @property
    def denominator(self) -> Integer:
        return Integer(self._den)

    def __int__(self) -> int:
        if self._den == 1:
            return self._num
        raise ValueError("Cannot convert rational number to integer")

    def __float__(self) -> float:
        return self._num / self._den

    def __complex__(self) -> complex:
        return float(self) + 0j

    def _to_same_type(self, other: Any) -> Rational:
        if isinstance(other, int):
            return Rational(other, 1)
        if isinstance(other, Integer):
            return Rational(int(other), 1)
        if isinstance(other, Rational):
            return other
        raise ValueError(f"Could not convert {other} to rational")

    def _add(self, other: Self) -> AbstractNumber:
        return Rational(
            self._num * other._den + self._den * other._num,
            self._den * other._den,
        )

    def _sub(self, other: Self) -> AbstractNumber:
        return Rational(
            self._num * other._den - self._den * other._num,
            self._den * other._den,
        )

    def _mul(self, other: Self) -> AbstractNumber:
        return Rational(
            self._num * other._num,
            self._den * other._den,
        )

    def _truediv(self, other: Self) -> AbstractNumber:
        return Rational(
            self._num * other._den,
            self._den * other._num,
        )

    def _pow(self, other: int) -> AbstractNumber:
        return Rational(
            self._num**other,
            self._den**other,
        )

    def _eq(self, other: Self) -> bool:
        return self._num == other._num and self._den == other._den

    def __abs__(self):
        return abs(float(self))


class GaussianInteger(AbstractNumber):
    """A Gaussian integer."""

    def __init__(self, re: int, im: int):
        """Initialise.

        Args:
            re: The real part
            im: The imaginary part
        """
        self._re = re
        self._im = im

    def __str__(self):
        return f"{self._re}+{self._im}j"

    def __repr__(self):
        return f"GaussianInteger({self._re}+{self._im}j)"

    @property
    def real(self) -> RealNumber:
        return Integer(self._re)

    @property
    def imag(self) -> RealNumber:
        return Integer(self._im)

    def conjugate(self) -> AbstractNumber:
        return GaussianInteger(self._re, -self._im)

    def as_latex(self) -> str:
        return f"{self._re}+{self._im}\\mathrm{{i}}"

    @property
    def numerator(self) -> AbstractNumber:
        return self

    @property
    def denominator(self) -> Integer:
        return Integer(1)

    def __int__(self) -> int:
        if self._im == 0:
            return self._re
        raise ValueError("Cannot convert complex number to integer")

    def __float__(self) -> float:
        if self._im == 0:
            return float(self._re)
        raise ValueError("Cannot convert complex number to integer")

    def __complex__(self) -> complex:
        return self._re + 1j * self._im

    def _to_same_type(self, other: Any) -> GaussianInteger:
        if isinstance(other, int):
            return GaussianInteger(other, 0)
        if isinstance(other, Integer):
            return GaussianInteger(int(other), 0)
        if isinstance(other, GaussianInteger):
            return other
        raise ValueError(f"Could not convert {other} to Gaussian integer")

    def _add(self, other: Self) -> AbstractNumber:
        return GaussianInteger(self._re + other._re, self._im + other._im)

    def _sub(self, other: Self) -> AbstractNumber:
        return GaussianInteger(self._re - other._re, self._im - other._im)

    def _mul(self, other: Self) -> AbstractNumber:
        return GaussianInteger(
            self._re * other._re - self._im * other._im,
            self._re * other._im + self._im * other._re,
        )

    def _truediv(self, other: Self) -> AbstractNumber:
        num = self * other.conjugate()
        den = other * other.conjugate()
        return GaussianRational(int(num.real), int(den), int(num.imag), int(den))

    def _mod(self, other: Self) -> AbstractNumber:
        num = self * other.conjugate()
        den = other * other.conjugate()
        return GaussianInteger(int(num.real % den), int(num.imag % den))

    def _floordiv(self, other: Self) -> AbstractNumber:
        num = self * other.conjugate()
        den = other * other.conjugate()
        return GaussianInteger(int(num.real // den), int(num.imag // den))

    def _eq(self, other: Self) -> bool:
        return self._re == other._re and self._im == other._im


class GaussianRational(AbstractNumber):
    """A Gaussian rational."""

    def __init__(self, re_numerator: int, re_denominator, im_numerator: int, im_denominator: int):
        """Initialise.

        Args:
            re_numerator: The numerator of the real part
            re_denominator: The denominator of the real part
            im_numerator: The numerator of the imaginary part
            im_denominator: The denominator of the imaginary part
        """
        if re_denominator < 0:
            re_numerator *= -1
            re_denominator *= -1
        if im_denominator < 0:
            im_numerator *= -1
            im_denominator *= -1
        hcf = math.gcd(abs(re_numerator), abs(re_denominator))
        self._re_num = re_numerator // hcf
        self._re_den = re_denominator // hcf
        hcf = math.gcd(abs(im_numerator), abs(im_denominator))
        self._im_num = im_numerator // hcf
        self._im_den = im_denominator // hcf

    def __str__(self):
        return f"{self._re_num}/{self._re_den} + {self._im_num}j/{self._im_den}"

    def __repr__(self):
        return f"GaussianRational({self})"

    @property
    def real(self) -> RealNumber:
        return Rational(self._re_num, self._re_den)

    @property
    def imag(self) -> RealNumber:
        return Rational(self._im_num, self._im_den)

    def as_latex(self) -> str:
        return f"\\frac{{{self._re_num}}}{{{self._re_den}}}+\\frac{{{self._im_num}}}{{{self._im_den}}}\\mathrm{{i}}"

    def conjugate(self) -> AbstractNumber:
        return GaussianRational(self._re_num, self._re_den, -self._im_num, self._im_den)

    @property
    def numerator(self) -> AbstractNumber:
        d = int(self.denominator)
        return GaussianInteger(
            self._re_num * d // self._re_den,
            self._im_num * d // self._im_den,
        )

    @property
    def denominator(self) -> Integer:
        return Integer(math.lcm(self._re_den, self._im_den))

    def __int__(self) -> int:
        if self._im_num == 0 and self._re_den == 1:
            return self._re_num
        raise ValueError("Cannot convert complex number to integer")

    def __float__(self) -> float:
        if self._im_num == 0:
            return self._re_num / self._re_den
        raise ValueError("Cannot convert complex number to integer")

    def __complex__(self) -> complex:
        return self._re_num / self._re_den + 1j * self._im_num / self._im_den

    def _to_same_type(self, other: Any) -> GaussianRational:
        if isinstance(other, int):
            return GaussianRational(other, 1, 0, 1)
        if isinstance(other, Integer):
            return GaussianRational(int(other), 1, 0, 1)
        if isinstance(other, Rational):
            return GaussianRational(int(other.numerator), int(other.denominator), 0, 1)
        if isinstance(other, GaussianInteger):
            return GaussianRational(int(other.real), 1, int(other.imag), 1)
        if isinstance(other, GaussianRational):
            return other
        raise ValueError(f"Could not convert {other} to Gaussian rational")

    def _add(self, other: Self) -> AbstractNumber:
        re = self.real + other.real
        im = self.imag + other.imag
        return GaussianRational(
            int(re.numerator), int(re.denominator), int(im.numerator), int(im.denominator)
        )

    def _sub(self, other: Self) -> AbstractNumber:
        re = self.real - other.real
        im = self.imag - other.imag
        return GaussianRational(
            int(re.numerator), int(re.denominator), int(im.numerator), int(im.denominator)
        )

    def _mul(self, other: Self) -> AbstractNumber:
        re = self.real * other.real - self.imag * other.imag
        im = self.real * other.imag + self.imag * other.real
        return GaussianRational(
            int(re.numerator), int(re.denominator), int(im.numerator), int(im.denominator)
        )

    def _truediv(self, other: Self) -> AbstractNumber:
        num = self * other.conjugate()
        den = other * other.conjugate()
        re = num.real / den
        im = num.imag / den
        return GaussianRational(
            int(re.numerator), int(re.denominator), int(im.numerator), int(im.denominator)
        )

    def _eq(self, other: Self) -> bool:
        return self.real == other.real and self.imag == other.imag


def _simplify_type(i: AbstractNumber) -> AbstractNumber:
    """Convert i to a simpler type if possible."""
    if isinstance(i, Rational):
        if i.denominator == 1:
            return _simplify_type(i.numerator)
    if isinstance(i, GaussianInteger):
        if i.imag == 0:
            return _simplify_type(i.real)
    if isinstance(i, GaussianRational):
        if i.imag == 0:
            return _simplify_type(i.real)
        if i.denominator == 1:
            return _simplify_type(i.numerator)
    return i


def _as_common_type(a: AbstractNumber, b: Any) -> tuple[AbstractNumber, AbstractNumber]:
    """Convert a and b to the same type."""
    try:
        return a, a._to_same_type(b)
    except ValueError:
        pass
    if isinstance(b, AbstractNumber):
        try:
            return b._to_same_type(a), b
        except ValueError:
            pass
    i = GaussianRational(1, 2, 1, 2)
    try:
        return i._to_same_type(a), i._to_same_type(b)
    except ValueError:
        raise ValueError(f"Could not find common type for {a} and {b}")


j = GaussianInteger(0, 1)
zero = Integer(0)
one = Integer(1)



"""Test documentation."""

import os

import pytest
import arctans


def test_vesion_numbers():
    if not os.path.isfile(
        os.path.join(os.path.dirname(os.path.realpath(__file__)), "../pyproject.toml")
    ):
        pytest.skip("Not running in source directory")
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../pyproject.toml")) as f:
        version = f.read().split('version = "')[1].split('"')[0]

    assert arctans.__version__ == version



"""Arctans."""

from arctans.arctans import arccotan, arctan
from arctans.numbers import Integer, Rational, GaussianInteger, GaussianRational, j, one, zero
from arctans.primes import is_irreducible
from arctans.reduction import reduce, convert_rational
from arctans.generation import generate

__version__ = "1.0.0"



import pytest
from arctans.numbers import Integer, Rational, GaussianInteger, GaussianRational
from utils import isclose

options = [Integer(2), Rational(1, 2), GaussianInteger(3, 4), GaussianRational(1, 2, 3, 4)]


@pytest.mark.parametrize("a", options)
@pytest.mark.parametrize("b", options)
def test_add(a, b):
    assert isclose(complex(a + b), complex(a) + complex(b))


@pytest.mark.parametrize("a", options)
@pytest.mark.parametrize("b", options)
def test_sub(a, b):
    assert isclose(complex(a - b), complex(a) - complex(b))


@pytest.mark.parametrize("a", options)
@pytest.mark.parametrize("b", options)
def test_mul(a, b):
    assert isclose(complex(a * b), complex(a) * complex(b))


@pytest.mark.parametrize("a", options)
@pytest.mark.parametrize("b", options)
def test_div(a, b):
    print(a, b)
    assert isclose(complex(a / b), complex(a) / complex(b))



import math
from arctans import arccotan, convert_rational, reduce
from utils import isclose


def test_convert_rational():
    a = convert_rational(arccotan(239))
    assert isclose(float(a), math.atan(1 / 239))


def test_machin_term2():
    a = reduce(arccotan(239))
    assert isclose(float(a), math.atan(1 / 239))



from math import pi
import pytest
from arctans import arctan, generate
from utils import isclose


@pytest.mark.parametrize("max_d", [10, 25])
def test_generate(max_d):
    formulae = generate([4 * arctan(1)], max_denominator=max_d, max_terms=8)
    for f in formulae:
        assert isclose(float(f), pi)
    print(f"Found {len(formulae)} new formulae for pi")
    formulae = generate(formulae, max_denominator=max_d, max_terms=8)
    for f in formulae:
        assert isclose(float(f), pi)
    print(f"Found {len(formulae)} more new formulae for pi")


def test_generate_with_numerator():
    formulae = generate([4 * arctan(1)], max_denominator=10, max_numerator=2, max_terms=8)
    for f in formulae:
        assert isclose(float(f), pi)
    print(f"Found {len(formulae)} formulae for pi")



from arctans import GaussianInteger
import pytest


def isclose(a, b, eps=1e-10):
    return abs(a - b) < eps


@pytest.mark.parametrize("a", [-3, 0, 2])
@pytest.mark.parametrize("b", [-3, 0, 2])
@pytest.mark.parametrize("c", [-3, 0, 2])
@pytest.mark.parametrize("d", [-3, 0, 2])
def test_add(a, b, c, d):
    w = GaussianInteger(a, b)
    z = GaussianInteger(c, d)
    assert isclose(complex(w + z), complex(w) + complex(z))


@pytest.mark.parametrize("a", [-3, 0, 2])
@pytest.mark.parametrize("b", [-3, 0, 2])
@pytest.mark.parametrize("c", [-3, 0, 2])
@pytest.mark.parametrize("d", [-3, 0, 2])
def test_sub(a, b, c, d):
    w = GaussianInteger(a, b)
    z = GaussianInteger(c, d)
    assert isclose(complex(w - z), complex(w) - complex(z))


@pytest.mark.parametrize("a", [-3, 0, 2])
@pytest.mark.parametrize("b", [-3, 0, 2])
@pytest.mark.parametrize("c", [-3, 0, 2])
@pytest.mark.parametrize("d", [-3, 0, 2])
def test_mult(a, b, c, d):
    w = GaussianInteger(a, b)
    z = GaussianInteger(c, d)
    assert isclose(complex(w * z), complex(w) * complex(z))


@pytest.mark.parametrize("a", [-3, 0, 2])
@pytest.mark.parametrize("b", [-3, 0, 2])
@pytest.mark.parametrize("c", [-3, 0, 2])
@pytest.mark.parametrize("d", [-3, 1, 2])
def test_mod(a, b, c, d):
    w = GaussianInteger(a, b)
    z = GaussianInteger(c, d)
    num = complex(w) * complex(z).conjugate()
    denom = int((complex(z) * complex(z).conjugate()).real)
    result = int(num.real) % denom + 1j * (int(num.imag) % denom)
    assert isclose(complex(w % z), result)


@pytest.mark.parametrize("a", [-3, 0, 2])
@pytest.mark.parametrize("b", [-3, 0, 2])
@pytest.mark.parametrize("c", [-3, 0, 2])
@pytest.mark.parametrize("d", [-3, 1, 2])
def test_division(a, b, c, d):
    w = GaussianInteger(a, b)
    z = GaussianInteger(c, d)
    assert isclose(complex(w / z), complex(w) / complex(z))


@pytest.mark.parametrize("a", [-3, 0, 2])
@pytest.mark.parametrize("b", [-3, 0, 2])
@pytest.mark.parametrize("c", [-3, 0, 2])
@pytest.mark.parametrize("d", [-3, 1, 2])
def test_integer_division(a, b, c, d):
    w = GaussianInteger(a, b)
    z = GaussianInteger(c, d)
    num = complex(w) * complex(z).conjugate()
    denom = int((complex(z) * complex(z).conjugate()).real)
    result = int(num.real) // denom + 1j * (int(num.imag) // denom)
    assert isclose(complex(w // z), result)


@pytest.mark.parametrize("a", [-3, 0, 2])
@pytest.mark.parametrize("b", [-3, 0, 2])
def test_abs(a, b):
    w = GaussianInteger(a, b)
    assert isclose(abs(w), abs(complex(w)))



"""Core base classes."""

from abc import ABC, abstractmethod


class Representable(ABC):
    """Object that can be represented as a string or as LaTeX."""

    @abstractmethod
    def __str__(self):
        pass

    @abstractmethod
    def __repr__(self):
        pass

    def __hash__(self):
        return hash(self.__repr__())

    @abstractmethod
    def as_latex(self) -> str:
        """Represent in LaTeX."""



import pytest
from arctans import arccotan, arctan, is_irreducible, reduce, convert_rational, Rational
from utils import isclose

reducible = [3, 7, 8, 13, 17, 18, 21]


@pytest.mark.parametrize("n", range(1, 23))
def test_irreducible(n):
    assert is_irreducible(n) == (n not in reducible)


@pytest.mark.parametrize("coefficient", range(1, 20))
@pytest.mark.parametrize("a", range(1, 20))
def test_convert_rational_integer(coefficient, a):
    arctan_a = coefficient * arctan(a)
    converted = convert_rational(arctan_a)
    assert isclose(float(arctan_a), float(converted))
    for i, j in converted.terms:
        assert j.numerator == 1
        assert i != 0


@pytest.mark.parametrize("numerator", range(1, 20))
@pytest.mark.parametrize("denominator", range(1, 20))
def test_convert_rational(numerator, denominator):
    arctan_n = arctan(Rational(numerator, denominator))
    converted = convert_rational(arctan_n)
    assert isclose(float(arctan_n), float(converted))
    for i, j in converted.terms:
        assert j.numerator == 1
        assert i != 0


@pytest.mark.parametrize("n", reducible)
@pytest.mark.parametrize("c", [-2, -1, 1, 3, Rational(1, 2)])
def test_reduction(c, n):
    arctan_n = c * arctan(n)
    print(c, arctan(n))
    print(arctan_n)
    reduced = reduce(arctan_n)
    assert isclose(float(arctan_n), float(reduced))
    assert reduced.nterms > 1


@pytest.mark.parametrize("n", reducible)
def test_reduction_leads_to_irreducible(n):
    arctan_n = arctan(n)
    reduced = reduce(arctan_n)
    for _, i in reduced.terms:
        assert i == 0 or i == 1 or is_irreducible(1 / i)


@pytest.mark.parametrize("n", range(1, 300))
def test_reduction_nonzero_coefficients(n):
    arctan_n = arccotan(n)
    reduced = reduce(arctan_n)
    assert isclose(float(arctan_n), float(reduced))
    for i, _ in reduced.terms:
        assert i != 0



"""Generation of new formulae."""

from arctans.arctans import arctan, AbstractTerm
from arctans.numbers import Rational
from arctans.reduction import reduce
from typing import Sequence


def generate(
    known_formula: AbstractTerm | Sequence[AbstractTerm],
    *,
    min_denominator: int = 1,
    max_denominator: int = 100,
    min_numerator: int = 1,
    max_numerator: int = 1,
    max_terms: int | None = None,
    max_coefficient_denominator: int | None = None,
    printing: bool = False,
) -> list[AbstractTerm]:
    """Generate new formulae.

    Args:
        known_formula: Known formula or formulae that all have the same value
        min_numerator: The minimum numerator to use for arctan arguments
        max_numerator: The maximum numerator to use for arctan arguments
        min_denominator: The minimum denominator to use for arctan arguments
        max_denominator: The maximum denominator to use for arctan arguments
        max_terms: The maximum number of arctan terms to include in the new formulae
        max_coefficient_denominator: The maximum allowbale denominator to use in the
            coefficients in the new formulae
        printing: Print information about progress

    Returns:
        A list of new formulae that have the same value as the known formula(e)
    """
    if isinstance(known_formula, AbstractTerm):
        value = float(known_formula)
        known_formulae: Sequence[AbstractTerm] = [known_formula]
    else:
        value = float(known_formula[0])
        for i in known_formula[1:]:
            assert abs(float(i) - value) < 0.0001
        known_formulae = known_formula
    new_formulae = []
    for denominator in range(min_denominator, max_denominator + 1):
        for numerator in range(min_numerator, max_numerator + 1):
            if printing:
                print(numerator, denominator)
            a = arctan(Rational(numerator, denominator))
            zero = reduce(a) - a
            for c, t in zero.terms:
                for f in known_formulae:
                    if t in f.term_dict:
                        new_f = f - zero * f.term_dict[t] / c
                        if new_f in known_formulae or new_f in new_formulae:
                            continue
                        if max_terms is not None and len(new_f.terms) > max_terms:
                            continue
                        if (
                            max_coefficient_denominator is not None
                            and max(c.denominator for c, a in new_f.terms)
                            > max_coefficient_denominator
                        ):
                            continue
                        new_formulae.append(new_f)
    return new_formulae



"""Test documentation."""

import os

import pytest

readme_snippets = []

if os.path.isfile(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../README.md")):
    with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../README.md")) as f:
        for part in f.read().split("```python")[1:]:
            readme_snippets.append(part.split("```")[0].strip())


@pytest.mark.parametrize("snippet", readme_snippets)
def test_readme_snippets(snippet):
    exec(snippet)



"""Mathematical utility functions."""

from arctans.numbers import GaussianInteger, Integer, j

primes = [2]


def pfactors(n: int) -> list[Integer]:
    """Get list of all prime factors of n.

    Args:
        n: An integer

    Returns:
        A list of the prime factors of n, including factors multiple times when they appear more than once in the prime factorisation
    """
    out = []

    p = Integer(2)
    while n > 1:
        while n % p == 0:
            out.append(p)
            n //= p
        p += 1

    return out


def largest_pfactor(n: int | Integer) -> Integer:
    """Compute the largest prime factor of n.

    Args:
        n: An integer

    Returns:
        The largest prime factor of n
    """
    if n < 2:
        raise ValueError(f"Cannot find largest prime factor of {n}")
    n = int(n)
    i = 2
    while i < n:
        if n % i == 0:
            n //= i
        else:
            i += 1
    return Integer(n)


def is_prime(n: Integer | int) -> bool:
    """Check if an integer is prime.

    Args:
        n: An integer

    Returns:
        True if n is prime
    """
    global primes

    i = primes[-1]
    while primes[-1] < n:
        for p in primes:
            if i % p == 0:
                break
        else:
            primes.append(i)
        i += 1
    return n in primes


def is_irreducible(n: Integer | int) -> bool:
    """Check if arctan(n) is irreducible.

    An arctan is irreducible iff it cannot be written as a
    weighted sum of integer arccotangents, or equivalently
    arctan(n) is irreducible iff the largest prime factor of
    1 + n**2 is greater than or equal to 2*n.

    Args:
        n: An integer

    Returns:
        True if n is irreducible
    """
    return largest_pfactor(1 + n**2) >= 2 * n


def is_gaussian_prime(n: GaussianInteger) -> bool:
    """Check if n is a Gaussian prime.

    Args:
        n: An integer

    Returns:
        True if n is a Gaussian prime
    """
    if n.imag == 0 or n.real == 0:
        k = abs(n.real) + abs(n.imag)
        return k % 4 == 3 and is_prime(k)
    return is_prime(n.real**2 + n.imag**2)


def is_gaussian_unit(n: GaussianInteger) -> bool:
    """Check if n is a Gaussian unit.

    Args:
        n: An integer

    Returns:
        True if n is 1, -1, i or -i
    """
    return n in [GaussianInteger(1, 0), GaussianInteger(-1, 0), j, -j]


def complex_factorise(
    n: GaussianInteger,
    istart: int = 0,
) -> list[GaussianInteger]:
    """Factorise a Gaussian integer into Gaussian primes.

    Args:
        n: An integer

    Returns:
        A list of Gaussian primes
    """
    if is_gaussian_unit(n) or is_gaussian_prime(n):
        return [n]
    lim = int(abs(n)) + 1
    for re in range(istart, lim + 1):
        for im in range(-lim, lim + 1):
            m = GaussianInteger(re, im)
            if abs(m) > 1 and is_gaussian_prime(m) and n % m == 0:
                return [m] + complex_factorise(n // m, re)
    raise RuntimeError(f"Could not fund factor of non-prime number: {n}")



"""Arctans."""

from arctans.numbers import AbstractNumber, Integer, RealNumber
from abc import abstractmethod
import math
from arctans.core import Representable


def _format_single_atan(a: AbstractNumber) -> str:
    """Format a single arctan."""
    if a.numerator == 1:
        return f"[{a.denominator}]"
    return f"arctan({a})"


class AbstractTerm(Representable):
    """Abstract term."""

    @property
    @abstractmethod
    def terms(self) -> list[tuple[AbstractNumber, AbstractNumber]]:
        """Return list of (coefficient, arctan) pairs."""

    @property
    @abstractmethod
    def term_dict(self) -> dict[AbstractNumber, AbstractNumber]:
        """Return dictionary {arctan: coefficient}."""

    def __eq__(self, other):
        if isinstance(other, AbstractTerm):
            return self.terms == other.terms
        return False

    def __neg__(self):
        return -1 * self

    def __add__(self, other):
        if isinstance(other, AbstractTerm):
            return ArctanSum(*self.terms, *other.terms)
        return NotImplemented

    def __iadd__(self, other):
        if not isinstance(other, AbstractTerm):
            return NotImplemented
        self = ArctanSum(*self.terms, *other.terms)
        return self

    def __sub__(self, other):
        if isinstance(other, AbstractTerm):
            return ArctanSum(*self.terms, *[(-i, j) for i, j in other.terms])
        else:
            return NotImplemented

    def __float__(self) -> float:
        out = 0.0
        for i, j in self.terms:
            out += float(i) * math.atan(float(j))
        return out

    def __mul__(self, other):
        return ArctanSum(*[(i * other, j) for i, j in self.terms])

    def __rmul__(self, other):
        return ArctanSum(*[(other * i, j) for i, j in self.terms])

    def __truediv__(self, other):
        return ArctanSum(*[(i / other, j) for i, j in self.terms])

    @property
    def nterms(self) -> int:
        """Number of terms."""
        return len(self.terms)

    def __hash__(self):
        return hash(self.__repr__())


class Zero(AbstractTerm):
    """Zero."""

    @property
    def terms(self) -> list[tuple[AbstractNumber, AbstractNumber]]:
        return []

    @property
    def term_dict(self) -> dict[AbstractNumber, AbstractNumber]:
        return {}

    def as_latex(self) -> str:
        return "0"

    def __str__(self) -> str:
        return "0"

    def __repr__(self) -> str:
        return "Zero()"

    def __float__(self) -> float:
        return 0.0


class Arctan(AbstractTerm):
    """A single arctan."""

    def __init__(self, arctan: AbstractNumber):
        """Initialise a single scaled arctan term.

        Args:
            coefficient: The coefficient that the arctan is scaled by
            arctan: The argument of the arctan

        """
        self._arctan = arctan

    @property
    def terms(self) -> list[tuple[AbstractNumber, AbstractNumber]]:
        return [(Integer(1), self._arctan)]

    @property
    def term_dict(self) -> dict[AbstractNumber, AbstractNumber]:
        return {self._arctan: Integer(1)}

    def as_latex(self) -> str:
        return f"\\arctan({self._arctan})"

    def __str__(self) -> str:
        return _format_single_atan(self._arctan)

    def __repr__(self) -> str:
        return f"Arctan({self._arctan})"


class ArctanSum(AbstractTerm):
    """The sum of some arctans."""

    def __init__(self, *terms: tuple[AbstractNumber, AbstractNumber]):
        """Initialise.

        Args:
            terms: A list of coefficient and arctan argument pairs
        """
        terms_dict: dict[AbstractNumber, AbstractNumber] = {}
        for c, a in terms:
            if isinstance(a, RealNumber) and a < 0:
                a *= -1
                c *= -1
            if a not in terms_dict:
                terms_dict[a] = Integer(0)
            terms_dict[a] += c
        self._terms = [(j, i) for i, j in terms_dict.items()]
        self._terms.sort(key=lambda i: 1 / abs(i[1]))
        self._terms = [i for i in self._terms if i[0] != 0]
        assert len(set([i[1] for i in self._terms])) == len([i[1] for i in self._terms])

    def as_latex(self) -> str:
        return " + ".join(
            ("" if i == 1 else f"{i}") + f"\\arctan({j})" for i, j in self._terms
        ).replace(
            "+ -",
            "- ",
        )

    def __repr__(self) -> str:
        return f"ArctanSum({self.__str__()})"

    def __str__(self) -> str:
        return " + ".join(f"{i}*{_format_single_atan(j)}" for i, j in self._terms).replace(
            "+ -",
            "- ",
        )

    @property
    def terms(self) -> list[tuple[AbstractNumber, AbstractNumber]]:
        return self._terms

    @property
    def term_dict(self) -> dict[AbstractNumber, AbstractNumber]:
        return {j: i for i, j in self._terms}


def arctan(a: AbstractNumber | int) -> AbstractTerm:
    """Symbolic arctangent.

    Args:
        a: The argument of the arctan

    Returns:
        arctan(a)

    """
    if isinstance(a, int):
        a = Integer(a)
    if isinstance(a, RealNumber) and a < 0:
        return -arctan(-a)
    if a == 0:
        return Zero()

    return Arctan(a)


def arccotan(a: AbstractNumber | int) -> AbstractTerm:
    """Symbolic arccotangent.

    Args:
        a: The argument of the arccotan

    Returns:
        arccotan(a)

    """
    if a == 0:
        return 2 * arctan(1)
    return arctan(Integer(1) / a)



import pytest
from arctans.primes import (
    largest_pfactor,
    complex_factorise,
    is_gaussian_prime,
)
from arctans import GaussianInteger

primes = [2]
for i in range(3, 1000, 2):
    for p in primes:
        if i % p == 0:
            break
    else:
        primes.append(i)


@pytest.mark.parametrize("n", range(2, 1000))
def test_largest_pfactor(n):
    assert largest_pfactor(n) == max(p for p in primes if n % p == 0)


@pytest.mark.parametrize("a", range(-13, 20))
@pytest.mark.parametrize("b", range(-13, 20))
def test_complex_primes(a, b):
    if abs(a) + abs(b) <= 1:
        # Is zero or a unit
        pytest.skip()
    lim = max(abs(a), abs(b))
    n = GaussianInteger(a, b)
    for i in range(-lim, lim + 1):
        for j in range(-lim, lim + 1):
            if a**2 + b**2 > i**2 + j**2 and abs(i) + abs(j) > 1 and n % GaussianInteger(i, j) == 0:
                assert not is_gaussian_prime(n)
                break
        else:
            continue
        break
    else:
        assert is_gaussian_prime(n)


@pytest.mark.parametrize("a", range(-13, 20))
@pytest.mark.parametrize("b", range(-13, 20))
def test_complex_factorise(a, b):
    if abs(a) + abs(b) <= 1:
        # Is zero or a unit
        pytest.skip()
    n = GaussianInteger(a, b)
    f = complex_factorise(n)
    if is_gaussian_prime(n):
        assert len(f) == 1 and f[0] == n
    else:
        assert len(f) > 1
        product = GaussianInteger(1, 0)
        for i in f:
            assert is_gaussian_prime(i)
            product *= i
        assert product == n



from math import pi
from arctans import arctan, arccotan
from arctans.arctans import AbstractTerm
from utils import isclose


def test_simplify():
    s = arctan(5) + 2 * arctan(5)
    assert s.nterms == 1
    assert s.terms[0] == (3, 5)


def test_machins_formula():
    s = 16 * arccotan(5) - 4 * arccotan(239)
    assert isclose(float(s), pi)


def test_add():
    a = 16 * arccotan(5)
    b = -4 * arccotan(239)

    assert isinstance(a + b, AbstractTerm)

    assert isclose(float(a + b), pi)


def test_sub():
    a = 16 * arccotan(5)
    b = 4 * arccotan(239)

    assert isinstance(a - b, AbstractTerm)

    assert isclose(float(a - b), pi)


def test_multiply():
    a = 4 * arccotan(5) - arccotan(239)

    assert isinstance(4 * a, AbstractTerm)
    assert isinstance(a * 4, AbstractTerm)

    assert isclose(float(4 * a), pi)
    assert isclose(float(a * 4), pi)


def test_division():
    a = 64 * arccotan(5) - 16 * arccotan(239)

    assert isinstance(a / 4, AbstractTerm)

    assert isclose(float(a / 4), pi)
